'''
$Log: bbslib.py,v $
Revision 1.1.1.1  2004/02/15 08:44:28  jangc
start

Revision 1.19  2004/02/14 22:04:05  pybbs
..

Revision 1.18  2004/01/18 04:31:44  pybbs
..

Revision 1.17  2004/01/08 01:21:58  pybbs
..

Revision 1.16  2004/01/04 13:49:35  pybbs
do_changeState is added
'''
import time
import re
from StringIO import StringIO
from urllib import quote
import cgi
import os
import pyConfigParser

import dblib
import config
from templateProcessor import templateProcessor, getTemplate
import linkProcessor

class InvaildOrder(Exception): pass

gdb = None
groups = None
globalNS = None

def connectDB():
    global gdb
    if not gdb:
        gdb = dblib.connectDB()

def disconnectDB():
    global gdb
    if gdb:
        dblib.closeDB(gdb)
        gdb = None

def escapeTextForDb(text):
    text = text.replace('\\', '\\\\')
    text = text.replace("'", "\\'")
    return text

def escapeWhiteSpace(text):
    sp = re.compile(' ( +)')
    def spacerepl(match):
        s = match.group()
        return ' '+'&nbsp;'*(len(s)-1)
    text = text.expandtabs()
    text = sp.sub(spacerepl, text)
    text = text.replace('\n', ' <br/>\n')
    return text

def escapeTagMarker(s):
    s = s.replace('<:', '&lt;:')
    return s.replace(':>', ':&gt;')

def escapeTextForm(body):
    body = body.replace('<br/>','\n')
    body = body.replace('&', '&amp;')
    body = escapeTagMarker(body)
    return body

# Schedule Manager
################################################################################
class ScheduleManager(list):
    def __init__(self, sortBy='', orderBy='', pageNo=1, where={}, conjunction='and'):
        self.sortBy = sortBy
        self.orderBy = orderBy
        self.pageNo = pageNo
        self.where = where
        self.conjunction = conjunction

        cursor = dblib.getCursor()
        query = 'SELECT * %s' % self.makeQuery()
        self.count = cursor.execute(query)
        while self:
            self.pop()
        for args in cursor.fetchall():
            self.append(Schedule(*args))
        return self.count

    def getLastPage(self):
        cursor = dblib.getCursor()
        query = 'SELECT COUNT(*) %s' % self.makeQuery()
        index = query.find('ORDER')
        query = query[:index]
        cursor.execute(query)
        self.totalCount = cursor.fetchone()[0]
        self.lastPage, mod = divmod(self.totalCount, config.pageSize)
        if mod:
            self.lastPage += 1
        self.lastPage = int(self.lastPage)
        return self.lastPage
        
    def makeQuery(self):
        sort = ''
        if self.sortBy and self.orderBy:
            sort = 'ORDER BY %s %s' % (self.sortBy, self.orderBy)
        where = []
        for key, valueList in self.where.items():
            if type(valueList) == type(''):
                valueList = [valueList]
            L = []
            for value in valueList:
                L.append('%s=%s' % (key, str(value)))
            where.append(' or '.join(L))
        where = filter(None, where)
        if where:
            where = 'where %s' % (' %s ' % self.conjunction).join(where)
        else:
            where = ''
         
        limit = ''
        if self.pageNo:
            limit = 'LIMIT %s, %s' % (config.pageSize * (self.pageNo-1), config.pageSize)
        self.queryOptions = 'FROM %s %s %s %s' % (config.tableName, where, sort, limit)
        file('sql', 'a').write(self.queryOptions)
        return self.queryOptions

    def getPageList(self):
        L = []
        for key, valueList in self.where.items():
            if type(valueList) == type(''):
                valueList = [valueList]
            for value in valueList:
                L.append('%s=%s' % (key, str(value)))
      
        aPage = '<a href="todo.py?sortBy=%s&orderBy=%s&pageNo=%s&%s">[%s]</a> '
        aPageList = []
        for N in range(1, self.getLastPage()+1):
            aPageList.append(aPage % (self.sortBy, self.orderBy, N, '&'.join(L), N))
        aPageList = ''.join(aPageList)
        return aPageList
        
    def reSort(self, sortBy='scheduleId', orderBy='desc'):
        try:
            if orderBy == 'desc':
                self.sort(lambda lhs, rhs: cmp(getattr(rhs, sortBy), getattr(lhs, sortBy)))
            elif orderBy == 'asc':
                self.sort(lambda lhs, rhs: cmp(getattr(lhs, sortBy), getattr(rhs, sortBy)))
            else:
               raise InvalidOrder
            return True
        except AttributeError, InvalidOrder:
            return False

    def toHTML(self):
        scheduleContents = map(lambda obj: obj.toHTML(sortBy=self.sortBy, orderBy=self.orderBy, pageNo=self.pageNo, where=self.where), self)
        scheduleContents = ''.join(scheduleContents)
        teulPath = os.path.join(config.skinPath, 'scheduleListHead.teul')
        templateHead = getTemplate(teulPath)
        teulPath = os.path.join(config.skinPath, 'scheduleListFoot.teul')
        templateFoot = getTemplate(teulPath)
        template = '%s%s%s' % (templateHead, scheduleContents, templateFoot)
        aPageList = self.getPageList()
        return templateProcessor(template, localNS=locals())
        
# Schedule
################################################################################
class Schedule:
    def __init__(self, scheduleId='NULL', importance=0, category=0, content='',date='',
                    deadline='', writer='tester', state=0):
        self.today = time.strftime('%Y-%m-%d %H:%M:%S') 
        self.scheduleId = scheduleId
        self.importance = importance
        self.category = category
        if date:
            self.date = str(date)
        else:
            self.date = time.strftime('%Y-%m-%d %H:%M:%S')
        if deadline:
            self.deadline = str(deadline)
        else:
            self.deadline = '0000-00-00 00:00:00'
        if len(self.deadline) != 19:               # FIXME
            self.deadline = self.deadline[:-3]
        self.writer = writer
        self.content = content
        self.state = state
        
    def escapeSubjectAndBody(self):
        self.escapedContentForDb = escapeTextForDb(self.content.rstrip())

    def escapeContentHtml(self):
        self.escapedHtmlContent = cgi.escape(self.content)
        self.escapedHtmlContent = escapeWhiteSpace(self.escapedHtmlContent)
        self.escapedHtmlContent = linkProcessor.linkProcessor(self.escapedHtmlContent)
        self.escapedHtmlContent = escapeTagMarker(self.escapedHtmlContent)

    def insert(self):
        self.scheduleId = 'NULL'
        cursor = dblib.getCursor()
        formatStringList = self.getFormatStringList()
        sqlArg = '(%s)' % (', '.join(formatStringList) % self.__dict__)
        query = "INSERT INTO %s VALUES %s" % (config.tableName, sqlArg)
        cursor.execute(query)
        self.scheduleId = cursor.insert_id()
        return self.scheduleId
    
    def update(self):
        def connection(field, formatStr):
            return '%s=%s' % (field, formatStr)
        cursor = dblib.getCursor()
        self.escapeSubjectAndBody()
        names = dblib.getFieldNames(config.tableName)
        sqlArg = ', '.join(map(connection, names, self.getFormatStringList()))
        where = 'scheduleId = %s' % self.scheduleId 
        query = "UPDATE %s SET %s WHERE %s" % (config.tableName, sqlArg % self.__dict__, where)
        cursor.execute(query)
            
    def toHTML(self, **kw):
        self.escapeContentHtml()
        state = getValueFromConfigInSkin('state', self.state)
        importance = getValueFromConfigInSkin('importance', self.importance)
        category = getValueFromConfigInSkin('category', self.category)
        content = self.escapedHtmlContent
        deadline = self.deadline
        daysLeft = getDaysLeft(self.today, self.deadline)
        imagePath = os.path.join(config.skinPath, 'images')
        aStateChangeButton = self.getStateChangeButton(**kw)
        teulPath = os.path.join(config.skinPath, 'scheduleListBody.teul')
        template = getTemplate(teulPath)
        return templateProcessor(template, globalNS=getGlobalNS(), localNS=locals())

    def getStateChangeButton(self, **kw):
        startLink = '<a href="todo.py?action=changeState&scheduleId=%s&cState=%s&' 
        etcList = []
        for key, value in kw.items():
            if key == 'where':
                continue
            etcList.append('%s=%s' % (key, value))
        etc = '&'.join(etcList)
        L = []
        for key, valueList in kw['where'].items():
            if type(valueList) == type(''):
                valueList = [valueList]
            for value in valueList:
                L.append('%s=%s' % (key, str(value)))
        if etc:
            startLink = startLink + '&' + etc + '&' + '&'.join(L)
        endLink = '"><img src="%s/%s" border="0"/></a> '
        if not hasattr(self, 'conf'):
            self.conf = pyConfigParser.CookedConfigParser()
            self.conf.read(os.path.join(config.skinPath, 'config.ini'))
        section = 'state'
        viewList, valueList = [], []
        linkList = []
        link = startLink + endLink
        for option in self.conf.options(section):
            imagePath = os.path.join(config.skinPath, 'images')
            optionValue = self.conf.get(section, option)
            linkList.append(link % (self.scheduleId, option, imagePath, optionValue))
        return ''.join(linkList)
                    
    def getFormatStringList(self):
        self.escapeSubjectAndBody()
        fieldInfo = dblib.getFieldInfo(config.tableName)
        fieldInfo = [f[:2] for f in fieldInfo]
        L = []
        for fieldName, fieldType in fieldInfo:
            if fieldName == 'content': fieldName = 'escapedContentForDb'
            fieldType = fieldType.lower()
            if fieldType.find('char') >= 0 or fieldType.find('datetime') >= 0 or fieldType.find('text') >= 0 :
                L.append("'%%(%s)s'" % fieldName)
            else:
                L.append('%%(%s)s' % fieldName)
        return L
        return '(%s)' % (', '.join(L) % self.__dict__)
    
def getSchedule(scheduleId):
    query = 'SELECT * FROM %s WHERE scheduleId=%s' % (config.tableName, scheduleId)
    cursor = dblib.getCursor()
    cursor.execute(query)
    args = cursor.fetchone()
    fieldNames = dblib.getFieldNames(config.tableName)
    try:
        schedule = Schedule(**dict(zip(fieldNames, args)))
        return schedule
    except TypeError:
        return None

def deleteSchedule(scheduleId):
    cursor = dblib.getCursor()
    ret = cursor.execute('DELETE FROM %s WHERE scheduleId=%s' % (config.tableName, scheduleId))
    return ret

def changeState(scheduleId, state):
    if not scheduleId:
        return False
    cursor = dblib.getCursor()
    query = "UPDATE %s SET state=%s WHERE scheduleId=%s" % (config.tableName, state, scheduleId)
    cursor.execute(query)
    return True


# following functions are used by skins. 
################################################################################
def getCookedDateFormat(date, format=None):
    if not date:
        return None
    if format == None:
        format = config.dateFormat
    import time
    try:
        ymd, hms= date.split()
        date = map(lambda field: int(field), ymd.split('-') + hms.split(':')) + [0] * 3
        return time.strftime(format, date)
    except TypeError:
        return ''
    except:
        return date

def strCutter(text, length=40, postfix="..."):
    if len(text) < length: return text
    pos = length - 1 
    char = text[pos] 
    while ord(char) > 127 and pos >= 0:
        pos -= 1
        char = text[pos]
    return text[:(length-pos) % 2 and length or length - 1] + postfix

def getValueFromConfigInSkin(section, option):
    cp = pyConfigParser.CookedConfigParser()
    cp.read(os.path.join(config.skinPath, 'config.ini'))
    return cp.get(section, str(option))

def getDaysLeft(today, deadline):
    # makeTimeStrToTuple
    today = str(getCookedDateFormat(today, '%Y %m %d')).split()
    deadline = str(getCookedDateFormat(deadline, '%Y %m %d')).split()
    today = map(int, today)
    deadline = map(int, deadline)
    today.extend([0] * 6)
    deadline.extend([0] * 6)
    today = tuple(today)
    deadline = tuple(deadline)
    # getDaysLeft
    secondsLeft = time.mktime(deadline) - time.mktime(today) 
    aDay = 60 * 60 * 24
    daysLeft, mod = divmod(secondsLeft, aDay)
    if mod > 0:
        daysLeft += 1
    return int(daysLeft)

def makeSelect(name, view=[], value=[], selected=''):
    selectStart = '<select name="%s">\n' % name
    selectEnd = '</select>\n'
    selectOptionList = []
    if value:
        selectOption = '    <option value="%s"%s>%s</option>'
        for i in range(len(view)):
            sel = selected == value[i] and ' selected' or ''
            selectOptionList.append(selectOption % (value[i], sel, view[i]))
    else:
        value = view
        selectOption = '    <option%s>%s</option>'
        for i in range(len(view)):
            sel = selected == view[i] and ' selected' or ''
            selectOptionList.append(selectOption % (sel, view[i]))
    selectOptions = '\n'.join(selectOptionList)
    return selectStart + selectOptions + selectEnd

def makeSelectFromConfigInSkin(name, section='', selected=''):
    if not section:
        section = name
    conf = pyConfigParser.CookedConfigParser()
    conf.read(os.path.join(config.skinPath, 'config.ini'))
    viewList, valueList = [], []
    for option in conf.options(section):
        viewList.append(conf.get(section, option))
        valueList.append(option)
    return makeSelect(name, viewList, valueList, selected)

def getGlobalNS():
    return {'strCutter':strCutter, 'getValueFromConfigInSkin':getValueFromConfigInSkin, 'getCookedDateFormat':getCookedDateFormat, 'getDaysLeft':getDaysLeft}
